// CBlocks : grid.h
//
// Datastructures and functions for grids
//
// See userguide.pdf for copyright information

#include <stdio.h>
#include <malloc.h>
#include <assert.h>
#include "misc.h"
#include "block.h"
#include "grid.h"

#undef MOVE_DIAGNOSE

#define MAX_MOVES 1000
#define OPPOSITE(x) ((x+2) % 4)

int x_lookup[4] =
{
	0,1,0,-1,
};

int y_lookup[4] =
{
	1,0,-1,0,
};


#ifdef MOVE_DIAGNOSE
extern FILE *diagnose;
#endif

/* Given an x and y coord, return the direction that they correspond to */
direction xy_to_direction(int x, int y)
{
  if (x==0 && y==1)
  {
    return north;
  }
  else if (x==0 && y==-1)
  {
    return south;
  }
  else if (x==1 && y==0)
  {
    return east;
  }
  else if (x==-1 && y==0)
  {
    return west;
  }
  else
  {
    return no_direction;
  }
}

static int get_move_no(void)
{
  static int current_move_no = 1;

  current_move_no = (current_move_no++) % MAX_MOVES;

#ifdef MOVE_DIAGNOSE
  fprintf(diagnose,"Current move no %d\n",current_move_no);
#endif

  return current_move_no;
}

/* Initialise a grid */
Grid *init_Grid(int width, int height)
{
  int i,j,k;
  Grid *g;

  g = (Grid *)malloc(sizeof(Grid));

  g->active=0;
  g->inactive=1;
  g->width = width;
  g->height = height;
  g->ab_x1 = g->ab_y1 = 0;
  g->ab_x2 = g->width-1;
  g->ab_y2 = g->height-1;

  for (i=0; i<2; i++)
  {
    /* Allocate space for the array of pointers to columns */
    g->the_grid[i] = (Grid_element ***)malloc(width*sizeof(Grid_element **));

    for (j=0; j<width; j++)
    {
      /* Allocate space for the components of the column - pointers to grid 
       * elements */
      g->the_grid[i][j] = (Grid_element **)malloc(height*sizeof(Grid_element *)); 

      for (k=0; k<height; k++)
      {
        /* Allocate space for the grid elements */
        g->the_grid[i][j][k] = (Grid_element *)malloc(sizeof(Grid_element));

        /* Initialise the element */
        init_Grid_element(g->the_grid[i][j][k]);
      }
    }
  }

  g->isselection = false;
  g->selx1 = g->sely1 = g->selx2 = g->sely2 = 0;
  g->keypress = false;

  return g;
}

/* Destroy a grid */
void destroy_Grid(Grid *g)
{
  int i,j,k;

  for (i=0; i<2; i++)
  {
    for (j=0; j<g->width; j++)
    {
      for (k=0; k<g->height; k++)
      {
        /* Free space from the grid elements */
        free(g->the_grid[i][j][k]);
      }
      /* Free space from the components of the column - pointers to grid 
       * elements */
      free(g->the_grid[i][j]);
    }

    /* Free space used by pointers to the active and inactive arrays */
    free(g->the_grid[i]);
  }

  free(g);
}

/* Copy the active array to the inactive array */
void copy_Grid(Grid *g)
{
  int i,j;

  for (i=g->ab_x1; i<=g->ab_x2; i++)
  {
    for (j=g->ab_y1; j<=g->ab_y2; j++)
    {
      copy_Grid_element(g->the_grid[g->inactive][i][j],g->the_grid[g->active][i][j]);
    }
  }
}

/* Return the type of block at x,y */
blocktype get_type(Grid *g, int x, int y)
{
  /* Range check */
  if (x>=0 && y>=0 && x<g->width && y<g->height)
  {
    return g->the_grid[g->active][x][y]->block.type;
  }
  else
  {
    /* If x,y is not in grid then report that block is empty */
    return empty;
  }
}

/* Return the orientation of the block at x,y */
direction get_orientation(Grid *g, int x, int y)
{
  /* Range check */
  if (x>=0 && y>=0 && x<g->width && y<g->height)
  {
    return g->the_grid[g->active][x][y]->block.orientation;
  }
  else
  {
    /* If x,y is not in grid then report that block facing north */
    return north;
  }
}

/* Return the direction that the block at x,y is going to move in.  Return
 * no_direction if it is not going to move */
direction get_move_dir(Grid *g, int x, int y)
{
  /* Range check */
  if (x>=0 && y>=0 && x<g->width && y<g->height)
  {
    Grid_element *ge = g->the_grid[g->active][x][y];

    if (ge->block_move.move_requested && !(ge->block_move.already_cancelled))
    {
      return g->the_grid[g->active][x][y]->block_move.dir;
    }
    else
    {
      return no_direction;
    }
  }
  else
  {
    /* If x,y is not in grid then report that block is not going to move */
    return no_direction;
  }
}

/* Return the connectivity of the block at x,y in direction dir */
Boolean get_connect(Grid *g, int x, int y, direction dir)
{
  /* Range check */
  if (x>=0 && y>=0 && x<g->width && y<g->height)
  {
    return g->the_grid[g->active][x][y]->block.connect[dir];
  }
  else
  {
    /* If x,y is not in grid then report that block is not connected */
    return False;
  }
}

/* Set the connectivity of the block at x,y in direction dir */
void set_connect(Grid *g, int x, int y, direction dir, Boolean connect)
{
  /* Range check */
  if (x>=0 && y>=0 && x<g->width && y<g->height)
  {
    g->the_grid[g->active][x][y]->block.connect[dir] = connect;
  }
  else
  {
    /* If x,y is not in grid then do nothing */
    ;
  }
}

/* Get the input to block x,y which comes from the block in direction dir */
int get_input(Grid *g, int x, int y, direction dir)
{
  x += x_lookup[dir];
  y += y_lookup[dir];

  /* Range check */
  if (x>=0 && y>=0 && x<g->width && y<g->height)
  {
    return g->the_grid[g->active][x][y]->block.output[OPPOSITE(dir)];
  }
  else
  {
    /* If x,y is not in grid then report that block outputs 0 */
    return 0;
  }
}

/* Set the output value of block x,y in direction dir */
void set_output(Grid *g, int x, int y, direction dir, int o)
{
  /* Range check */
  if (x>=0 && y>=0 && x<g->width && y<g->height)
  {
    g->the_grid[g->active][x][y]->block.output[dir] = o;
  }
  else
  {
    /* If x,y is not in grid then do nothing */
    ;
  }
}

/* Cancel a previously requested move */
static void cancel_move(Grid *g, int x, int y)
{

#ifdef MOVE_DIAGNOSE
  fprintf(diagnose,"Cancelling at %d,%d\n",x,y);
#endif

  /* Range check */
  if (x>=0 && y>=0 && x<g->width && y<g->height)
  {
    Grid_element *ge = g->the_grid[g->active][x][y];

    if (x != NULL_COORD && y != NULL_COORD)
    {
      if (!(ge->block_move.already_cancelled))
      {
        /* Store coords of block in move direction because they are used 
         * a lot */
        int xo = x+ge->block_move.xo, yo = y+ge->block_move.yo;
        direction d;

        ge->block_move.already_cancelled = True;

        if (get_type(g,xo,yo) != empty)
        {
          cancel_move(g,xo,yo);
        }
        else
        {
          /* Block at xo,yo is empty so xo,yo must be in range */

          /* Since this move is about to be cancelled, the empty block's
           * move_requested flag should be set */
          assert(g->the_grid[g->active][xo][yo]->block_move.move_requested);

          g->the_grid[g->active][xo][yo]->block_move.requester_x = NULL_COORD;
          g->the_grid[g->active][xo][yo]->block_move.requester_y = NULL_COORD;
        }

        /* Now cancel the moves of connected blocks */
        for (d=north; d<=west; d=(direction)(d+1))
        {
          if (ge->block.connect[d] && d != ge->block_move.dir)
          {
            cancel_move(g,x+x_lookup[d],y+y_lookup[d]);
          }
        }
      }
    }
  }
}

/* See if a block is able to move in direction dir */
static Boolean able_to_move(Grid *g, int x, int y, direction dir, int move_no)
{
#ifdef MOVE_DIAGNOSE
  fprintf(diagnose,"Seeing if %d,%d is able to move in dir %d\n",x,y,dir);
#endif

  /* Range check */
  if (x>=0 && y>=0 && x<g->width && y<g->height)
  {
    /* Store a pointer to this element because it is used a lot */
    Grid_element *ge = g->the_grid[g->active][x][y];

	  
	if (ge->block_move.asked_if_able == move_no)
    {
      /* If a block has already been asked if it is able to move, return
       * true because if it was unable to move, it will already have returned
       * false and the move will not proceed */

#ifdef MOVE_DIAGNOSE
      fprintf(diagnose,"Yes - already asked\n");
#endif

      return True;
    }
    else
    {
      ge->block_move.asked_if_able = move_no;

      if (ge->block.type == empty)
      {
        if (ge->block_move.move_requested)
        {
          /* If empty block is already going to be moved into, cancel
           * the move request which led to the block moving in and 
           * return False */

#ifdef MOVE_DIAGNOSE
           fprintf(diagnose,"No - empty block with something coming\n");
#endif

           cancel_move(g,ge->block_move.requester_x,
                       ge->block_move.requester_y);

           return False;
        }
        else
        {
          /* Block is available so return true */

#ifdef MOVE_DIAGNOSE
          fprintf(diagnose,"Yes - empty block available\n");
#endif

          return True;
        }
      }
      else
      {
        if (!(ge->block_move.already_cancelled))
        {
          if (ge->block_move.move_requested)
          {
            /* If this block is already going to move, only return true if it
             * is going to move in the same direction as this request.  If it 
             * is not moving in the same direction, cancel the original 
             * request to move it and return False */
            if (ge->block_move.xo == x_lookup[dir]  &&  
                ge->block_move.yo == y_lookup[dir])
            {

#ifdef MOVE_DIAGNOSE
              fprintf(diagnose,"Yes - already going in same dir\n");
#endif

              return True;
            }
            else
            {

#ifdef MOVE_DIAGNOSE
              fprintf(diagnose,"Cancelling requester of %d,%d due to different requested dirs",x,y);
#endif

              cancel_move(g,ge->block_move.requester_x,
                          ge->block_move.requester_y);
              return False;
            }
          }
          else
          {
            direction d;
            Boolean neighbours_able = True;
            
            /* See if connected blocks are able to move */
            for (d=north; d<=west && neighbours_able; d=(direction)(d+1))
            {
              /* No need to check block in direction dir because this
               * has already been done a few lines above */
              if (d == dir || ge->block.connect[d])
              {
                neighbours_able = (Boolean)(neighbours_able && able_to_move(g,x+x_lookup[d],y+y_lookup[d],dir,move_no));
              }
            }

#ifdef MOVE_DIAGNOSE
            fprintf(diagnose,"Yes - all neighbours able\n");
#endif

            return neighbours_able;
          }
        }
        else
        {

#ifdef MOVE_DIAGNOSE
          fprintf(diagnose,"No - already been moved and cancelled\n");
#endif

          /* This block has already been moved and cancelled */
          return False;
        }
      }
    }
  }
  else
  {
    /* If x,y is not in grid then report that it cannot move */
    return False;
  }
}

/* Ask a block to move in direction dir */
static void request_move(Grid *g, int x, int y, int requester_x, int requester_y,
                  direction dir)
{

#ifdef MOVE_DIAGNOSE
  fprintf(diagnose,"Asking %d,%d to move in dir %d\n",x,y,dir);
#endif
  
  /* Theoretically, a range check is not necessary because able_to_move
   * should have been called just before this function.  To help 
   * debugging, range is checked anyway */
  if (x>=0 && y>=0 && x<g->width && y<g->height)
  {  
    direction d;
    Grid_element *ge = g->the_grid[g->active][x][y];
    
    /* Check that this block hasn't already been asked to move */
    if (!ge->block_move.move_requested)
    {
      /* Check that this block isn't empty */
      if (ge->block.type != empty)
      {
        /* Indicate that a move has been requested */
        ge->block_move.move_requested = True;

        /* Direction in which block is to move */
        ge->block_move.dir = dir;
        ge->block_move.xo = x_lookup[dir];
        ge->block_move.yo = y_lookup[dir];

#ifdef MOVE_DIAGNOSE
  fprintf(diagnose,"Block %d,%d set to move in dir %d\n",x,y,dir);
#endif

        /* Store co-ords of the block which originally initiated this set of 
         * moves */
        ge->block_move.requester_x = requester_x;
        ge->block_move.requester_y = requester_y;

        for (d=north; d<=west; d=(direction)(d+1))
        {
          if (d == dir || ge->block.connect[d])
          {
            request_move(g,x+x_lookup[d],y+y_lookup[d],requester_x,requester_y,
                         dir);
          }
        }
      }
      else
      {
        ge->block_move.move_requested = True;
        ge->block_move.requester_x = requester_x;
        ge->block_move.requester_y = requester_y;
      }
    }
    else
    {
      /* Because this block has already been asked to move, the conflict
       * resolution strategies should have ensured that it is moving in
       * the same direction as is requested */
       assert(ge->block_move.dir == dir);
    }
  }
  else
  {
    /* Should never reach here */
    assert(False);
  }
}

/* Try to move the block at x,y in the direction dir */
void move_block(Grid *g, int x, int y, int requester_x, int requester_y, direction dir)
{
  if (get_type(g,x,y) != empty)
  {
    if (able_to_move(g,x,y,dir,get_move_no()))
    {
      request_move(g,x,y,requester_x,requester_y,dir);
    }  
  }
}

/* Alter the status of the new_connect flag for direction dir of the
 * block at x,y */
static void set_new_connect(Grid *g, int x, int y, direction dir, Boolean connect)
{
  /* Range check */
  if (x>=0 && y>=0 && x<g->width && y<g->height)
  {
    g->the_grid[g->active][x][y]->block_move.new_connect[dir] = connect;
  }
}

/* Reset the status of the new_connect flag for direction dir of the
 * block at x,y */
static void reset_new_connect(Grid *g, int x, int y, direction dir)
{
  /* Range check */
  if (x>=0 && y>=0 && x<g->width && y<g->height)
  {
    g->the_grid[g->active][x][y]->block_move.new_connect[dir] = 
      g->the_grid[g->active][x][y]->block.connect[dir];
  }
}

/* Alter the status of the connect_change flag for direction dir of the
 * block at x,y */
static void set_connect_change(Grid *g, int x, int y, direction dir, Boolean connect)
{
  /* Range check */
  if (x>=0 && y>=0 && x<g->width && y<g->height)
  {
    g->the_grid[g->active][x][y]->block_move.connect_change[dir] = connect;
  }
}

/* Try to fuse/unfuse the block at x,y with the block in direction dir */
void fuse_blocks(Grid *g, int x, int y, direction dir, Boolean connect)
{
  /* Range check */
  if (x>=0 && y>=0 && x<g->width && y<g->height)
  {
    Grid_element *ge = g->the_grid[g->active][x][y];

    if ((ge->block.type != empty) && 
       (get_type(g,x+x_lookup[dir],y+y_lookup[dir]) != empty))
    {
      if (!(ge->block_move.connect_change[dir]))
      {
        ge->block_move.connect_change[dir] = True;
        ge->block_move.new_connect[dir] = connect;

        set_connect_change(g,x+x_lookup[dir],y+y_lookup[dir],(direction)(OPPOSITE(dir)),True);
        set_new_connect(g,x+x_lookup[dir],y+y_lookup[dir],(direction)(OPPOSITE(dir)),connect);
      }
      else
      {
        if (ge->block_move.new_connect[dir] != connect)
        {
          /* There is a fuse/unfuse conflict so set the new_connect value back
           * to it's original state */
          ge->block_move.new_connect[dir] = ge->block.connect[dir];

          /* Do the same for the neighbour block */
          reset_new_connect(g,x+x_lookup[dir],y+y_lookup[dir],(direction)(OPPOSITE(dir)));
        }
      }
    }
  }
}

/* Try to rotate the block at x,y */
void rotate(Grid *g, int x, int y)
{
  /* Range check */
  if (x>=0 && y>=0 && x<g->width && y<g->height)
  {
    Grid_element *ge = g->the_grid[g->active][x][y];

    if (!(ge->block_move.rotated))
    {
      ge->block_move.rotated = True;
    }
  }
}

void place_block(Grid *g, int x, int y, blocktype bt, direction dir)
{
	Grid_element *ge = g->the_grid[g->active][x][y];
    direction d;

    ge->block.type = bt;
    ge->block.orientation = dir;

    for (d=north; d<=west; d=(direction)(d+1))
    {
      ge->block.output[d]=0;
    }

	if (bt == empty)
	{
	  int xd,yd;
      for (d=north; d<=west; d=(direction)(d+1))
	  {
        ge->block.connect[d]=False;

		xd = x + x_lookup[d];
		yd = y + y_lookup[d];

       if (xd>=0 && yd>=0 && xd<g->width && yd<g->height)
	   {
		 g->the_grid[g->active][xd][yd]->block.connect[OPPOSITE(d)] = False;
	   }

	  }
	}


    if (x<g->ab_x1)
		g->ab_x1 = x;
	if (x>g->ab_x2)
		g->ab_x2 = x;
	if (y<g->ab_y1)
	    g->ab_y1 = y;
	if (y>g->ab_y2)
	    g->ab_y2 = y;
}

/* Try to create a new block at x,y */
void create_new_block(Grid *g, int x, int y, blocktype bt, direction dir)
{
  /* Range check */
  if (x>=0 && y>=0 && x<g->width && y<g->height)
  {
    Grid_element *ge = g->the_grid[g->active][x][y];

    if (ge->block.type == empty)
    {

      if (!(ge->block_move.create_block))
      {

        if (ge->block_move.move_requested)
        {

#ifdef MOVE_DIAGNOSE
          fprintf(diagnose,"Calling cancel in create\n");
#endif

          /* another block is about to move in, so cancel it and don't
           * create anything here */
          cancel_move(g,ge->block_move.requester_x,ge->block_move.requester_y);
          ge->block_move.create_block = True;
          ge->block_move.new_block = empty;
        }
        else
        {
          ge->block_move.create_block = True;
          ge->block_move.new_block = bt;
          ge->block_move.new_block_dir = dir;

		  if (x<g->ab_x1)
			  g->ab_x1 = x;
		  if (x>g->ab_x2)
			  g->ab_x2 = x;
		  if (y<g->ab_y1)
			  g->ab_y1 = y;
		  if (y>g->ab_y2)
			  g->ab_y2 = y;

        }
      }
      else
      {
        /* Two creators want to create a block here, so don't let either
         * of them do it */
        ge->block_move.new_block = empty;
      }
    }
  }
}

void make_selection(Grid *g, int x1, int y1, int x2, int y2)
{
	g->isselection = true;

	g->selx1 = x1;
	g->selx2 = x2;
	g->sely1 = y1;
	g->sely2 = y2;
}

