// CBlocks : process.h
//
// Functions for processing a grid
//
// See userguide.pdf for copyright information

#include <assert.h>
#include <stdio.h>
#include "misc.h"
#include "process.h"
#include "grid.h"
#include "block.h"

#undef MOVE_DIAGNOSE

extern int x_lookup[4],y_lookup[4];
extern void (*block_funcs[NUM_BLOCKS])(Grid *g, int x, int y);

#ifdef MOVE_DIAGNOSE
extern FILE *diagnose;
#endif

/* Process the whole of the active array */
void process_Grid(Grid *g)
{
  int x,y,i;
  int minx,miny,maxx,maxy;

  minx = g->width-1;
  miny = g->height-1;
  maxx = 0;
  maxy = 0;

  copy_Grid(g);

  for (x=g->ab_x1; x<=g->ab_x2; x++)
  {
    for (y=g->ab_y1; y<=g->ab_y2; y++)
    {
      i = g->the_grid[g->active][x][y]->block.type;
      if (i)
      {
        block_funcs[i](g,x,y);

        if (x<minx)
          minx = x;
        if (y<miny)
          miny = y;
        if (x>maxx)
          maxx = x;
        if (y>maxy)
          maxy = y;
      }
    }
  }

  perform_fuse_rotate_create(g);
  perform_moves(g);

  g->active = g->inactive;
  g->inactive = 1 - g->inactive;

  g->ab_x1 = (minx ? minx - 1:0);
  g->ab_y1 = (miny ? miny - 1:0);
  g->ab_x2 = ((maxx == g->width-1)?maxx:maxx+1);
  g->ab_y2 = ((maxy == g->height-1)?maxy:maxy+1);
}

/* Deal with fusing & rotating & creating*/
void perform_fuse_rotate_create(Grid *g)
{
  int x,y;
  direction dir;

  for (x=g->ab_x1; x<=g->ab_x2; x++)
  {
    for (y=g->ab_y1; y<=g->ab_y2; y++)
    {
      for (dir=north; dir<=west; dir=(direction)(dir+1))
      {
        if (g->the_grid[g->active][x][y]->block.type != empty)
        {
          check_fuse_move_conflict(g,x,y,dir);
          perform_rotate(g,x,y);
        }
        else
        {
          perform_create(g,x,y);
        }
      }
    }
  }
}

/* Move any blocks which need moving */
void perform_moves(Grid *g)
{
  int x,y,xo,yo;
  Grid_element *tmp,*previous;
  direction move_dir;

  for (x=g->ab_x1; x<=g->ab_x2; x++)
  {
    for (y=g->ab_y1; y<=g->ab_y2; y++)
    {
      xo = 0;
      yo = 0;
      previous = NULL;

      if (((move_dir = g->the_grid[g->active][x][y]->block_move.dir) != no_direction)
         && (!g->the_grid[g->active][x][y]->block_move.already_cancelled))
      {

#ifdef MOVE_DIAGNOSE
        fprintf(diagnose,"Looking at %d,%d\n",x,y);
#endif
        
        do
        {
          g->the_grid[g->active][x+xo][y+yo]->block_move.dir = no_direction;
          g->the_grid[g->active][x+xo][y+yo]->block_move.xo = 0;
          g->the_grid[g->active][x+xo][y+yo]->block_move.yo = 0;

#ifdef MOVE_DIAGNOSE
          fprintf(diagnose,"Moving %d,%d\n",x+xo,y+yo);
#endif

          tmp = g->the_grid[g->inactive][x+xo][y+yo];
          g->the_grid[g->inactive][x+xo][y+yo] = previous;
          previous = tmp;

          xo = xo+x_lookup[move_dir];
          yo = yo+y_lookup[move_dir];

        } while (g->the_grid[g->active][x+xo][y+yo]->block_move.dir != no_direction);

        g->the_grid[g->inactive][x][y] = g->the_grid[g->inactive][x+xo][y+yo];
        g->the_grid[g->inactive][x+xo][y+yo] = previous;
      }
#ifdef MOVE_DIAGNOSE
      else
      {
        if (g->the_grid[g->active][x][y]->block_move.already_cancelled)
        {
          fprintf(diagnose,"Alread cancelled at %d,%d\n",x,y);
        }
      }
#endif

    }
  }
}

/* Ensure that there is not going to be a fuse/move conflict & perform
 * a fuse/unfuse if necessary */
void check_fuse_move_conflict(Grid *g, int x, int y,direction dir)
{
  if (g->the_grid[g->active][x][y]->block_move.connect_change[dir]) 
  {
    if (g->the_grid[g->active][x][y]->block_move.new_connect[dir])
    {
      /* If two blocks are about to connect, make sure they are not
       * going to move in different directions.  If they are, stop
       * the connection from going ahead */
      if (get_move_dir(g,x,y) != 
          get_move_dir(g,x+x_lookup[dir],y+y_lookup[dir]))
      {
        /* Only set this blocks new connectivity to false.  The other
         * one will be dealt with separately */
        g->the_grid[g->active][x][y]->block_move.new_connect[dir] = False;
      }
    }

    g->the_grid[g->inactive][x][y]->block.connect[dir] =
      g->the_grid[g->active][x][y]->block_move.new_connect[dir];
  }
}

/* Rotate any blocks which need rotating */
void perform_rotate(Grid *g, int x, int y)
{
  if (g->the_grid[g->active][x][y]->block_move.rotated)
  {
    direction d;
    int prev = g->the_grid[g->active][x][y]->block.output[west];

    for (d=north; d<=west; d=(direction)(d+1))
    {
      g->the_grid[g->inactive][x][y]->block.output[d] = prev;
      prev = g->the_grid[g->active][x][y]->block.output[d];
    }

    g->the_grid[g->inactive][x][y]->block.orientation = (direction)
      ((g->the_grid[g->active][x][y]->block.orientation + 1) % 4);
  }
}

/* Create any blocks which need creating */
void perform_create(Grid *g, int x, int y)
{
  if (g->the_grid[g->active][x][y]->block_move.create_block &&
      (g->the_grid[g->active][x][y]->block_move.new_block != empty))
  {
    g->the_grid[g->inactive][x][y]->block.type =
      g->the_grid[g->active][x][y]->block_move.new_block;

    g->the_grid[g->inactive][x][y]->block.orientation =
      g->the_grid[g->active][x][y]->block_move.new_block_dir;

    /* This is a newly created block, so it won't be connected to anything
     * and none of it's outputs will be active */
  }
}
